# Copyright 2013 Answers for AWS LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from cache import Memorize
from exceptions import *

import boto
from boto import ec2

import time

__all__ = ('GraffitiMonkey', 'Logging')
log = logging.getLogger(__name__)


class GraffitiMonkey(object):
    def __init__(self, region, profile, instance_tags_to_propagate, volume_tags_to_propagate, dryrun, append):
        # This list of tags associated with an EC2 instance to propagate to
        # attached EBS volumes
        self._instance_tags_to_propagate = instance_tags_to_propagate

        # This is a list of tags associated with a volume to propagate to
        # a snapshot created from the volume
        self._volume_tags_to_propagate = volume_tags_to_propagate

        # The region to operate in
        self._region = region

        # The profile to use
        self._profile = profile

        # Whether this is a dryrun
        self._dryrun = dryrun

        # If we are appending tags
        self._append = append

        log.info("Connecting to region %s using profile %s", self._region, self._profile)
        try:
            self._conn = ec2.connect_to_region(self._region, profile_name=self._profile)
        except boto.exception.NoAuthHandlerFound:
            raise GraffitiMonkeyException('No AWS credentials found - check your credentials')
        except boto.provider.ProfileNotFoundError:
            log.info("Connecting to region %s using default credentials", self._region)
            try:
                self._conn = ec2.connect_to_region(self._region)
            except boto.exception.NoAuthHandlerFound:
                raise GraffitiMonkeyException('No AWS credentials found - check your credentials')




    def propagate_tags(self):
        ''' Propagates tags by copying them from EC2 instance to EBS volume, and
        then to snapshot '''

        self.tag_volumes()
        self.tag_snapshots()


    def tag_volumes(self):
        ''' Gets a list of all volumes, and then loops through them tagging
        them '''

        log.info('Getting list of all volumes')
        volumes = self._conn.get_all_volumes()
        total_vols = len(volumes)
        log.info('Found %d volumes', total_vols)
        this_vol = 0
        for volume in volumes:
            this_vol +=1
            log.info ('Processing volume %d of %d total volumes', this_vol, total_vols)
            if volume.status != 'in-use':
                log.debug('Skipping %s as it is not attached to an EC2 instance, so there is nothing to propagate', volume.id)
                continue
            for attempt in range(5):
                try:
                    self.tag_volume(volume)
                except boto.exception.EC2ResponseError, e:
                    log.error("Encountered Error %s on volume %s", e.error_code, volume.id)
                    break
                except boto.exception.BotoServerError, e:
                    log.error("Encountered Error %s on volume %s, waiting %d seconds then retrying", e.error_code, volume.id, attempt)
                    time.sleep(attempt)
                else:
                    break
            else:
                log.error("Encountered Error %s on volume %s, %d retries failed, continuing", e.error_code, volume.id, attempt)
                continue

        log.info('Completed processing all volumes')


    def tag_volume(self, volume):
        ''' Tags a specific volume '''

        instance_id = None
        if volume.attach_data.instance_id:
            instance_id = volume.attach_data.instance_id
        device = None
        if volume.attach_data.device:
            device = volume.attach_data.device

        instance_tags = self._get_resource_tags(instance_id)

        tags_to_set = {}
        if self._append:
            tags_to_set = self._get_resource_tags(volume.id)
        for tag_name in self._instance_tags_to_propagate:
            log.debug('Trying to propagate instance tag: %s', tag_name)
            if tag_name in instance_tags:
                value = instance_tags[tag_name]
                tags_to_set[tag_name] = value

        # Additional tags
        tags_to_set['instance_id'] = instance_id
        tags_to_set['device'] = device

        if self._dryrun:
            log.info('DRYRUN: Volume %s would have been tagged %s', volume.id, tags_to_set)
        else:
            self._set_resource_tags(volume, tags_to_set)
        return True


    def tag_snapshots(self):
        ''' Gets a list of all snapshots, and then loops through them tagging
        them '''

        log.info('Getting list of all snapshots')
        snapshots = self._conn.get_all_snapshots(owner='self')
        total_snaps = len(snapshots)
        log.info('Found %d snapshots', total_snaps)
        this_snap = 0
        for snapshot in snapshots:
            this_snap +=1
            log.info ('Processing snapshot %d of %d total snapshots', this_snap, total_snaps)
            for attempt in range(5):
                try:
                    self.tag_snapshot(snapshot)
                except boto.exception.EC2ResponseError, e:
                    log.error("Encountered Error %s on snapshot %s", e.error_code, snapshot.id)
                    break
                except boto.exception.BotoServerError, e:
                    log.error("Encountered Error %s on snapshot %s, waiting %d seconds then retrying", e.error_code, snapshot.id, attempt)
                    time.sleep(attempt)
                else:
                    break
            else:
                log.error("Encountered Error %s on snapshot %s, %d retries failed, continuing", e.error_code, snapshot.id, attempt)
                continue
        log.info('Completed processing all snapshots')

    def tag_snapshot(self, snapshot):
        ''' Tags a specific snapshot '''

        volume_id = snapshot.volume_id
#        if volume_id == '':
#            log.debug('Skipping %s as it does not have volume information', snapshot.id)
#            continue

        volume_tags = self._get_resource_tags(volume_id)

        tags_to_set = {}
        if self._append:
            tags_to_set = self._get_resource_tags(snapshot.id)
        for tag_name in self._volume_tags_to_propagate:
            log.debug('Trying to propagate volume tag: %s', tag_name)
            if tag_name in volume_tags:
                value = volume_tags[tag_name]
                tags_to_set[tag_name] = value


        if self._dryrun:
            log.info('DRYRUN: Snapshot %s would have been tagged %s', snapshot.id, tags_to_set)
        else:
            self._set_resource_tags(snapshot, tags_to_set)
        return True


    @Memorize
    def _get_resource_tags(self, resource_id):
        ''' Gets all of the tags associated with an AWS resource (such as an
        instance-id, volume-id, etc) as a dictionary '''

        resource_tags = {}
        if resource_id:
            # Get the set of tags for a volume
            log.debug('Fetching tags for %s', resource_id)
            tags = self._conn.get_all_tags({'resource-id': resource_id})
            for tag in tags:
                resource_tags[tag.name] = tag.value

        return resource_tags


    def _set_resource_tags(self, resource, tags):
        ''' Sets the tags on the given AWS resource '''

        if not isinstance(resource, ec2.ec2object.TaggedEC2Object):
            msg = 'Resource %s is not an instance of TaggedEC2Object' % resource
            raise GraffitiMonkeyException(msg)

        for tag_key, tag_value in tags.iteritems():
            if not tag_key in resource.tags or resource.tags[tag_key] != tag_value:
                log.info('Tagging %s with [%s: %s]', resource.id, tag_key, tag_value)
                resource.add_tag(tag_key, tag_value)



class Logging(object):
    # Logging formats
    _log_simple_format = '%(asctime)s [%(levelname)s] %(message)s'
    _log_detailed_format = '%(asctime)s [%(levelname)s] [%(name)s(%(lineno)s):%(funcName)s] %(message)s'

    def configure(self, verbosity = None):
        ''' Configure the logging format and verbosity '''

        # Configure our logging output
        if verbosity >= 2:
            logging.basicConfig(level=logging.DEBUG, format=self._log_detailed_format, datefmt='%F %T')
        elif verbosity >= 1:
            logging.basicConfig(level=logging.INFO, format=self._log_detailed_format, datefmt='%F %T')
        else:
            logging.basicConfig(level=logging.INFO, format=self._log_simple_format, datefmt='%F %T')

        # Configure Boto's logging output
        if verbosity >= 4:
            logging.getLogger('boto').setLevel(logging.DEBUG)
        elif verbosity >= 3:
            logging.getLogger('boto').setLevel(logging.INFO)
        else:
            logging.getLogger('boto').setLevel(logging.CRITICAL)



# Copyright 2013 Answers for AWS LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import sys

from graffiti_monkey.core import GraffitiMonkey, Logging
from graffiti_monkey import __version__
from graffiti_monkey.exceptions import GraffitiMonkeyException

from boto.utils import get_instance_metadata


__all__ = ('run', )
log = logging.getLogger(__name__)


class GraffitiMonkeyCli(object):
    def __init__(self):
        self.region = None
        self.profile = None
        self.monkey = None
        self.args = None
        self.config = {"_instance_tags_to_propagate": ['Name'],
                       "_volume_tags_to_propagate": ['Name', 'instance_id', 'device']}
        self.dryrun = False
        self.append = False

    @staticmethod
    def _fail(message="Unknown failure", code=1):
        log.error(message)
        sys.exit(code)

    def get_argv(self):
        """
        The parse_args method from ArgumentParser expects to not get the script title when arguments are passed to the
        method. So the first element is omitted.
        """
        return sys.argv[1:]

    def set_cli_args(self):
        parser = argparse.ArgumentParser(description='Propagates tags from AWS EC2 instances to EBS volumes, and then to EBS snapshots. This makes it much easier to find things down the road.')
        parser.add_argument('--region', metavar='REGION',
                            help='the region to tag things in (default is current region of EC2 instance this is running on). E.g. us-east-1')
        parser.add_argument('--profile', metavar='PROFILE',
                            help='the profile (credentials) to use to connect to EC2')
        parser.add_argument('--verbose', '-v', action='count',
                            help='enable verbose output (-vvv for more)')
        parser.add_argument('--version', action='version', version='%(prog)s ' + __version__,
                            help='display version number and exit')
        parser.add_argument('--config', '-c', nargs="?", type=argparse.FileType('r'),
                        default=None, help="Give a yaml configuration file")
        parser.add_argument('--dryrun', action='store_true',
                            help='dryrun only, display tagging actions but do not perform them')
        parser.add_argument('--append', action='store_true',
                            help='append propagated tags to existing tags (up to a total of ten tags)')
        self.args = parser.parse_args(self.get_argv())

    @staticmethod
    def fail_due_to_bad_config_file(self):
        self._fail("Something went wrong reading the passed yaml config file. "
                          "Make sure to use valid yaml syntax. "
                          "Also the start of the file should not be marked with '---'.", 6)

    def set_config(self):
        if self.args.config:
            try:
                import yaml
            except:
                log.error("When the config parameter is used, you need to have the python PyYAML library.")
                log.error("It can be installed with pip `pip install PyYAML`.")
                sys.exit(5)

            try:
                #TODO: take default values and these can be overwritten by config
                self.config = yaml.load(self.args.config)
                if self.config is None:
                    self.fail_due_to_bad_config_file()
            except:
                self.fail_due_to_bad_config_file()



    def set_region(self):
        if "region" in self.config.keys():
            self.region = self.config["region"]
        elif self.args.region:
            self.region = self.args.region
        else:
            # If no region was specified, assume this is running on an EC2 instance
            # and work out what region it is in
            log.debug("Figure out which region I am running in...")
            instance_metadata = get_instance_metadata(timeout=5)
            log.debug('Instance meta-data: %s', instance_metadata)
            if not instance_metadata:
                GraffitiMonkeyCli._fail('Could not determine region. This script is either not running on an EC2 instance (in which case you should use the --region option), or the meta-data service is down')

            self.region = instance_metadata['placement']['availability-zone'][:-1]
        log.debug("Running in region: %s", self.region)

    def set_profile(self):
        if "profile" in self.config.keys():
            self.profile = self.config["profile"]
        elif self.args.profile:
            self.profile = self.args.profile
        else:
            self.profile = 'default'
        log.debug("Using profile: %s", self.profile)

    def set_dryrun(self):
        self.dryrun = self.args.dryrun

    def set_append(self):
        self.append = self.args.append

    def initialize_monkey(self):
        self.monkey = GraffitiMonkey(self.region,
                                     self.profile,
                                     self.config["_instance_tags_to_propagate"],
                                     self.config["_volume_tags_to_propagate"],
                                     self.dryrun,
                                     self.append)


    def start_tags_propagation(self):
        self.monkey.propagate_tags()

    def exit_succesfully(self):
        log.info('Graffiti Monkey completed successfully!')
        sys.exit(0)

    def run(self):
        self.set_cli_args()

        Logging().configure(self.args.verbose)
        log.debug("CLI parse args: %s", self.args)

        self.set_config()
        self.set_region()
        self.set_profile()
        self.set_dryrun()
        self.set_append()

        try:
            self.initialize_monkey()
            self.start_tags_propagation()

        except GraffitiMonkeyException as e:
            GraffitiMonkeyCli._fail(e.message)

        self.exit_succesfully()


def run():
    cli = GraffitiMonkeyCli()
    cli.run()
